#include "stdafx.h"
#include "EllipsePoint.h"
#include <complex>
#include <math.h>
#include <float.h>

namespace {
	// -1`1Nv
	double clamp( double v ) {
		if ( v < -1.0 ) {
			v = -1.0;
		}
		else if ( v > 1.0 ) {
			v = 1.0;
		}
		return v;
	}

	// 2̉Zo
	// a,b,c : W
	// solves: 
	void solv2Equation( const std::complex< double > &a, const std::complex< double > &b, const std::complex< double > &c, std::complex< double > solves[ 2 ] ) {
		solves[ 0 ] = ( -b + sqrt( b * b - 4.0 * a * c ) ) / ( 2.0 * a );
		solves[ 1 ] = ( -b - sqrt( b * b - 4.0 * a * c ) ) / ( 2.0 * a );
	}

	// f̗Zo
	std::complex<double> cubicRoot( const std::complex<double> &val ) {
		double d = pow( abs( val ), 1.0 / 3.0 );
		double r = arg( val ) / 3.0;
		return std::complex<double>( d * cos( r ), d * sin( r ) );
	}

	// ZoꂽtŒZƂȂ_PZo
	bool calcMinDistPoint( double x, double y, double a, double b, int solveNum, const std::complex< double > *ts, double eps, double &px, double &py, double &dist ) {
		dist = FLT_MAX;
		for ( int i = 0; i < solveNum; ++i ) {
			// ݂̂Iʂ
			const std::complex< double > t = ts[ i ];
			if ( abs( t.imag() ) > eps || abs( a + t * b ) < eps || abs( b + t * a ) < eps ) {
				continue;
			}
			double cos = clamp( x / ( a + t.real() * b ) );
			double sin = clamp( y / ( b + t.real() * a ) );
			if ( abs( cos ) > abs( sin ) ) {
				sin = sqrt( 1.0 - cos * cos ) * ( sin >= 0.0 ? 1.0 : -1.0 );
			}
			else {
				cos = sqrt( 1.0 - sin * sin ) * ( cos >= 0.0 ? 1.0 : -1.0 );
			}
			double nx = b * cos * t.real();
			double  ny = a * sin * t.real();
			double L = sqrt( nx * nx + ny * ny );
			if ( L < dist ) {
				dist = L;
				px = a * cos;
				py = b * sin;
			}
		}
		return ( dist != FLT_MAX );
	}
}




// ȉ~Ɠ_Ƃ̋Zo
//  xLen : ȉ~X̒
//  yLen : ȉ~Y̒
//  x,y  : _̍W
//  dist : ŒZ
//  px,py: ŋߐړ_
// ߂l: ŒZ܂ꍇtrueAŒZ݂ȂȂǕs̏ꍇfalse
bool distEllipsePoint( double xLen, double yLen, double x, double y, double &dist, double &px, double &py ) {

	xLen = ( xLen < 0.0 ? 0.0 : xLen );
	yLen = ( yLen < 0.0 ? 0.0 : yLen );

	// ȉ~kނĂꍇ͐Ƃ̋ɂȂ
	if ( xLen == 0.0 ) {
		double dy = abs( y ) - yLen;
		if ( dy > 0.0 ) {
			dist = sqrt( x * x + dy * dy );
			return true;
		}
		dist = abs( x );
		return true;
	}
	if ( yLen == 0.0 ) {
		double dx = abs( x ) - xLen;
		if ( dx > 0.0 ) {
			dist = sqrt( y * y + dx * dx );
			return true;
		}
		dist = abs( y );
		return true;
	}

	// p[^i}yP̋LQƂ̎j
	const double a = xLen;
	const double b = yLen;
	const double a2 = a * a;
	const double b2 = b * b;
	const double x2 = x * x;
	const double y2 = y * y;
	const double A = 2.0 / ( a * b ) * ( a2 + b2 );
	const double B = 4.0 + ( a2 - x2 ) / b2 + ( b2 - y2 ) / a2;
	const double C = 2.0 / ( a * b ) * ( a2 + b2 - x2 - y2 );
	const double D = 1.0 - x2 / a2 - y2 / b2;
	const double E = A / 4.0;
	const double p = B - 6.0 * E * E;
	const double q = C - 2.0 * B * E + 8.0 * E * E * E;
	const double r = D - C * E + B * E * E - 3.0 * E * E * E * E;
	const double alpha = 2.0 * p;
	const double beta = p * p - 4.0 * r;
	const double gamma = -q * q;
	const double f = beta - alpha * alpha / 3.0;
	const double g = gamma - alpha * beta / 3.0 + 2.0 / 27.0 * alpha * alpha * alpha;

	static const std::complex< double > w1 = std::complex< double >( -0.5, sqrt( 3.0 ) / 2.0 );		// x^3 = 1̋1
	static const std::complex< double > w2 = std::complex< double >( -0.5, -sqrt( 3.0 ) / 2.0 );	// x^3 = 1̋2
	static const double eps = 0.00000000001;

	// q=0̎̓P[X
	//  A~̏ꍇ
	if ( abs( q ) < eps ) {
		int solveNum = 4;
		std::complex<double> h[ 2 ] = {};
		solv2Equation( 1.0, p, r, h );
		std::complex<double> ts[ 4 ] = {};
		ts[ 0 ] = sqrt( h[ 0 ] ) - E;
		ts[ 1 ] = -sqrt( h[ 0 ] ) - E;
		ts[ 2 ] = sqrt( h[ 0 ] ) - E;
		ts[ 3 ] = -sqrt( h[ 0 ] ) - E;
		
		// ŋߐړ_Zo
		return calcMinDistPoint( x, y, a, b, solveNum, ts, eps, px, py, dist );
	}

	std::complex< double > v[ 3 ];	// p[^v̉

	const double adjD = -4.0 * f * f * f - 27.0 * g * g;	// ʎ
	if ( adjD >= 0.0 ) {
		// ݂̂Ȃ̂ŃrG^̌p
		double m = sqrt( -f / 3.0 );
		double n = -g / m / m;
		for ( int j = 0; j < 3; ++j ) {
			double np2m = clamp( n / ( 2.0 * m ) );
			v[ j ]._Val[ 0 ] = 2.0 * m * cos( 1.0 / 3.0 * acos( np2m ) + j * 2.0 / 3.0 * 3.14159265358979323846 );
		}
	}
	else {
		// 1A2Ȃ̂ŃJ_m̌p
		std::complex< double > rt = sqrt( pow( g / 2.0, 2.0 ) + pow( f / 3.0, 3 ) );
		std::complex< double > L = std::complex< double >( -g / 2.0, +0.0 ) + rt;
		std::complex< double > R = std::complex< double >( -g / 2.0, -0.0 ) - rt;	// +0.0, -0.0̕LcubicRoot֐ŉ]߂̂ŏdvI
		std::complex< double > CL = cubicRoot( L );
		std::complex< double > CR = cubicRoot( R );
		v[ 0 ] = CL + CR;
		v[ 1 ] = w1 * CL + w2 * CR;
		v[ 2 ] = w2 * CL + w1 * CR;
	}

	// 4Ė@xNg̃XP[ƂȂtZo
	std::complex< double > ts[ 12 ] = {};
	int solveNum = 0;
	for ( int i = 0; i < 3; ++i ) {
		std::complex< double > solves[ 4 ];
		std::complex< double > u = v[ i ] - alpha / 3.0;
		solv2Equation( 1.0, -sqrt( u ), ( p + u ) / 2.0 + q / ( 2.0 * sqrt( u ) ), solves );
		solv2Equation( 1.0f, sqrt( u ), ( p + u ) / 2.0 - q / ( 2.0 * sqrt( u ) ), solves + 2 );
		ts[ solveNum ] = solves[ 0 ] - E;
		ts[ solveNum + 1 ] = solves[ 1 ] - E;
		ts[ solveNum + 2 ] = solves[ 2 ] - E;
		ts[ solveNum + 3 ] = solves[ 3 ] - E;
		solveNum += 4;
	}

	// ŋߐړ_Zo
	return calcMinDistPoint( x, y, a, b, solveNum, ts, eps, px, py, dist );
}